// $Id: ComObject.cpp,v 1.27 2002/02/15 22:14:50 cthuang Exp $
#pragma warning(disable: 4786)
#include <stdarg.h>
#include <stdexcept>
#include "CommandProcessor.h"
#include "naCmd.h"
#include "objectCmd.h"
#include "refCmd.h"
#include "ComModule.h"
#include "InterfaceAdapter.h"
#include "ComObject.h"

// prefix prepended to operation name for property get
static const char getPrefix[] = "_get_";

// prefix prepended to operation name for property put
static const char setPrefix[] = "_set_";

ComObject::ComObject (const Class::Interfaces &interfaces,
                      Tcl_ThreadId threadId,
                      Tcl_Interp *interp,
                      TclObject servant,
                      TclObject destructor,
                      bool isSink):
    m_refCount(0),
    m_defaultInterface(*(interfaces.front())),
    m_threadId(threadId),
    m_interp(interp),
    m_servant(servant),
    m_destructor(destructor),
    m_supportErrorInfo(*this),
    m_pDispatch(0),
    m_isSink(isSink)
{
//    Tcl_Preserve(reinterpret_cast<ClientData>(m_interp));
    ComModule::instance().lock();

    for (Class::Interfaces::const_iterator p = interfaces.begin();
     p != interfaces.end(); ++p) {
        const Interface *pInterface = *p;
        m_supportedInterfaceMap.insert(
            pInterface->iid(), const_cast<Interface *>(pInterface));
    }

    m_pDefaultAdapter = implementInterface(m_defaultInterface);
}

ComObject::~ComObject ()
{
    if (m_registeredActiveObject) {
        // TODO: This call may return an error but I don't want to throw an
        // exception from a destructor.
        RevokeActiveObject(m_activeObjectHandle, 0);
    }

    m_iidToAdapterMap.forEach(Delete());
    delete m_pDispatch;

    // Execute destructor Tcl command if defined.
    int length;
    Tcl_GetStringFromObj(m_destructor, &length);
    if (length > 0) {
        TclObject script(m_destructor);
        script.lappend(m_servant);
        eval(script);
    }

    ComModule::instance().unlock();
//    Tcl_Release(reinterpret_cast<ClientData>(m_interp));
}

void
ComObject::registerActiveObject (REFCLSID clsid)
{
    HRESULT hr = RegisterActiveObject(
        unknown(), clsid, ACTIVEOBJECT_WEAK, &m_activeObjectHandle);
    if (FAILED(hr)) {
        _com_issue_error(hr);
    }
    m_registeredActiveObject = true;
}

InterfaceAdapter *
ComObject::implementInterface (const Interface &interfaceDesc)
{
    InterfaceAdapter *pAdapter = new InterfaceAdapter(*this, interfaceDesc);
    m_iidToAdapterMap.insert(interfaceDesc.iid(), pAdapter);
    return pAdapter;
}

ComObject *
ComObject::newInstance (
    const Interface &defaultInterface,
    Tcl_ThreadId threadId,
    Tcl_Interp *interp,
    TclObject servant,
    TclObject destructor,
    bool isSink)
{
    Class::Interfaces interfaces;
    interfaces.push_back(&defaultInterface);

    return new ComObject(
	interfaces,
	threadId,
	interp,
	servant,
	destructor,
	isSink);
}

ComObject *
ComObject::newInstance (
    const Class::Interfaces &interfaces,
    Tcl_ThreadId threadId,
    Tcl_Interp *interp,
    TclObject servant,
    TclObject destructor,
    bool isSink)
{
    ComObject *pComObject = new ComObject(
	interfaces,
	threadId,
	interp,
	servant,
	destructor,
	isSink);
    return pComObject;
}

int
ComObject::eval (TclObject script, TclObject *pResult)
{
    return CommandProcessor::instance().eval(
        m_threadId, m_interp, script, pResult);
}

// Implement IUnknown methods

HRESULT
ComObject::queryInterface (REFIID iid, void **ppvObj)
{
    if (IsEqualIID(iid, IID_IUnknown)) {
	*ppvObj = m_pDefaultAdapter;
        addRef();
	return S_OK;
    }

    if (IsEqualIID(iid, IID_IDispatch)) {
        // Expose the operations of the default interface through IDispatch.
        if (m_pDispatch == 0) {
            m_pDispatch = new InterfaceAdapter(*this, m_defaultInterface, true);
        }
	*ppvObj = m_pDispatch;
        addRef();
	return S_OK;
    }

    if (IsEqualIID(iid, IID_ISupportErrorInfo)) {
	*ppvObj = &m_supportErrorInfo;
        addRef();
	return S_OK;
    }

    InterfaceAdapter *pAdapter = m_iidToAdapterMap.find(iid);
    if (pAdapter == 0) {
        const Interface *pInterface = m_supportedInterfaceMap.find(iid);
        if (pInterface != 0) {
            pAdapter = implementInterface(*pInterface);
        }
    }

    if (pAdapter != 0) {
	*ppvObj = pAdapter;
        addRef();
	return S_OK;
    }

    *ppvObj = 0;
    return E_NOINTERFACE;
}

ULONG
ComObject::addRef ()
{
    InterlockedIncrement(&m_refCount);
    return m_refCount;
}

ULONG
ComObject::release ()
{
    InterlockedDecrement(&m_refCount);
    if (m_refCount == 0) {
	delete this;
        return 0;
    }
    return m_refCount;
}

// Generate a name for a Tcl variable used to hold an argument out value.

static TclObject
getOutVariableName (const Parameter &param)
{
    return TclObject("::tcom::arg_" + param.name());
}

// Convert IDispatch argument to Tcl value.

static TclObject
getArgument (Tcl_ThreadId threadId,
             Tcl_Interp *interp,
             VARIANT *pArg,
             const Parameter &param)
{
    if (vtMissing == pArg) {
        return newNaObj();

    } else if (param.flags() & PARAMFLAG_FOUT) {
        // Get name of Tcl variable to hold out value.
        TclObject varName = getOutVariableName(param);

        if (param.flags() & PARAMFLAG_FIN) {
            // For in/out parameters, set the Tcl variable to the input value.
            TclObject value(pArg, param.type(), interp);

            // Construct a Tcl command to set the variable.
            TclObject script(Tcl_NewListObj(0, 0));
            script.lappend(Tcl_NewStringObj("set", -1));
            script.lappend(varName);
            script.lappend(value);

            CommandProcessor::instance().eval(threadId, interp, script);
        }
        return varName;

    } else {
        return TclObject(pArg, param.type(), interp);
    }
}

// Fill exception information structure.

static void
fillExcepInfo (EXCEPINFO *pExcepInfo,
               HRESULT hresult,
               const char *source,
               const char *description)
{
    if (pExcepInfo != 0) {
        memset(pExcepInfo, 0, sizeof(EXCEPINFO));
        pExcepInfo->scode = hresult;

        _bstr_t bstrSource(source);
        pExcepInfo->bstrSource = SysAllocString(bstrSource);

        if (description != 0) {
            _bstr_t bstrDescription(description);
            pExcepInfo->bstrDescription = SysAllocString(bstrDescription);
        }
    }
}

static void
putOutVariant (Tcl_Interp *interp,
               VARIANT *pDest,
               TclObject &tclObject,
               const Type &type)
{
    switch (type.vartype()) {
    case VT_BOOL:
        *V_BOOLREF(pDest) = tclObject.getBool() ? VARIANT_TRUE : VARIANT_FALSE;
        break;

    case VT_R4:
        *V_R4REF(pDest) = tclObject.getDouble();
        break;

    case VT_R8:
        *V_R8REF(pDest) = tclObject.getDouble();
        break;

    case VT_DISPATCH:
    case VT_UNKNOWN:
    case VT_USERDEFINED:
        {
            IUnknown *pUnknown;

            Tcl_Obj *pObj = tclObject;
            if (pObj->typePtr == &unknownPointerType) {
                pUnknown = reinterpret_cast<IUnknown *>(
                    pObj->internalRep.otherValuePtr);
            } else {
                Reference *pRef = referenceHandleSupport.find(tclObject);
                pUnknown = (pRef == 0) ? 0 : pRef->unknown();
            }

            *V_UNKNOWNREF(pDest) = pUnknown;

            // The COM rules say we must increment the reference count of
            // interface pointers returned from methods.
            if (pUnknown != 0) {
                pUnknown->AddRef();
            }
        }
        break;

    case VT_BSTR:
        {
#if TCL_MINOR_VERSION >= 2
            // Uses Unicode function introduced in Tcl 8.2.
            *V_BSTRREF(pDest) = SysAllocString(
		static_cast<const wchar_t *>(tclObject.getUnicode()));
#else
            *V_BSTRREF(pDest) = SysAllocString(tclObject.c_str());
#endif
        }
        break;

    case VT_VARIANT:
        {
            // Must increment reference count of interface pointers returned
            // from methods.
            tclObject.toVariant(
                V_VARIANTREF(pDest), Type::variant(), interp, true);
        }
        break;

    default:
        *V_I4REF(pDest) = tclObject.getLong();
    }
}

HRESULT
ComObject::invoke (InterfaceAdapter *pAdapter,
                   DISPID dispid,
                   REFIID /*riid*/,
                   LCID /*lcid*/,
                   WORD wFlags,
                   DISPPARAMS *pDispParams,
                   VARIANT *pVarResult,
                   EXCEPINFO *pExcepInfo,
                   UINT *pArgErr)
{
    // Get the method description for method being invoked.
    const Method *pMethod = pAdapter->findDispatchMethod(dispid);
    if (pMethod == 0) {
        return DISP_E_MEMBERNOTFOUND;
    }

    HRESULT hresult;

    try {
        // Get the method or property to invoke on the servant.
        std::string operation;
        if (wFlags & DISPATCH_METHOD) {
            operation = pMethod->name();

        } else if (wFlags & DISPATCH_PROPERTYGET) {
            operation = getPrefix + pMethod->name();

        } else if (wFlags & (DISPATCH_PROPERTYPUT | DISPATCH_PROPERTYPUTREF)) {
            operation = setPrefix + pMethod->name();

        } else {
            return DISP_E_MEMBERNOTFOUND;
        }

        // Construct Tcl command call.
        TclObject script(m_servant);
        script.lappend(
            Tcl_NewStringObj(const_cast<char *>(operation.c_str()), -1));

        // Set the argument error pointer in case we need to use it.
        UINT argErr;
        if (pArgErr == 0) {
            pArgErr = &argErr;
        }

        // Convert arguments to Tcl values.
        // TODO: Should handle named arguments differently than positional
        // arguments.
        const Method::Parameters &parameters = pMethod->parameters();

        int argIndex = pDispParams->cArgs - 1;
        for (unsigned paramIndex = 0; paramIndex < pDispParams->cArgs;
         ++paramIndex, --argIndex) {
            const Parameter &param = parameters[paramIndex];

            // Append argument value.
            VARIANT *pArg = &(pDispParams->rgvarg[argIndex]);
            try {
                script.lappend(
                    getArgument(m_threadId, m_interp, pArg, param));
            }
            catch (_com_error &) {
                *pArgErr = argIndex;
                throw;
            }
        }

        // Call the Tcl command.
        TclObject result;
        int completionCode = eval(script, m_isSink ? 0 : &result);
        if (completionCode != TCL_OK) {
            fillExcepInfo(
                pExcepInfo,
                E_UNEXPECTED,
                m_servant.c_str(),
                result.c_str());
            return DISP_E_EXCEPTION;
        }

        // Get values of out parameters.
        int i = 0;
        for (Method::Parameters::const_iterator p = parameters.begin();
         p != parameters.end(); ++p, ++i) {
            if (p->flags() & PARAMFLAG_FOUT) {
                // Get name of Tcl variable that holds out value.
                TclObject varName = getOutVariableName(*p);

                // Construct a Tcl command to get the variable value.
                TclObject script(Tcl_NewListObj(0, 0));
                script.lappend(Tcl_NewStringObj("set", -1));
                script.lappend(varName);

                TclObject value;
                int completionCode = eval(script, &value);

                if (completionCode == TCL_OK) {
                    putOutVariant(
                        m_interp,
                        &pDispParams->rgvarg[pDispParams->cArgs - 1 - i],
                        value,
                        p->type());
                }
            }
        }

        // Convert return value.
        if (pMethod->type().vartype() != VT_VOID) {
            // Must increment reference count of interface pointers returned
            // from methods.
            result.toVariant(pVarResult, pMethod->type(), 0, true);
        }

        hresult = S_OK;
    }
    catch (_com_error &e) {
        fillExcepInfo(pExcepInfo, e.Error(), m_servant.c_str(), 0);
        hresult = DISP_E_EXCEPTION;
    }
    return hresult;
}

// Convert the native value that the va_list points to into a Tcl object.
// Returns a va_list pointing to the next argument.

static va_list
convertNativeToTclObject (va_list pArg,
                          Tcl_Interp *interp,
                          TclObject &tclObject,
                          const Type &type,
                          bool byRef=false)
{
    switch (type.vartype()) {
    case VT_BOOL:
        tclObject = Tcl_NewBooleanObj(
            byRef ? *va_arg(pArg, VARIANT_BOOL *) : va_arg(pArg, VARIANT_BOOL));
        break;

    case VT_DATE:
    case VT_R4:
    case VT_R8:
        tclObject = Tcl_NewDoubleObj(
            byRef ? *va_arg(pArg, double *) : va_arg(pArg, double));
        break;

    case VT_DISPATCH:
    case VT_UNKNOWN:
    case VT_USERDEFINED:
        {
            IUnknown *pUnknown = va_arg(pArg, IUnknown *);
            if (pUnknown == 0) {
                tclObject = Tcl_NewObj();
            } else {
                const Interface *pInterface =
                    InterfaceManager::instance().find(type.iid());
                tclObject = referenceHandleSupport.newObj(
                    interp, Reference::newReference(pUnknown, pInterface));
            }
        }
        break;

    case VT_NULL:
        tclObject = Tcl_NewObj();
        break;

    case VT_LPWSTR:
    case VT_BSTR:
        {
#if TCL_MINOR_VERSION >= 2
            // Uses Unicode function introduced in Tcl 8.2.
            Tcl_UniChar *pUnicode = va_arg(pArg, Tcl_UniChar *);
            if (pUnicode != 0) {
                tclObject = Tcl_NewUnicodeObj(pUnicode, -1);
            } else {
                tclObject = Tcl_NewObj();
            }
#else
            _bstr_t str(va_arg(pArg, wchar_t *));
            tclObject = Tcl_NewStringObj(str, -1);
#endif
        }
        break;

    default:
        tclObject = Tcl_NewLongObj(
            byRef ? *va_arg(pArg, int *) : va_arg(pArg, int));
    }

    return pArg;
}

// Convert the native value that the va_list points to into a Tcl object.
// Returns a va_list pointing to the next argument.

static va_list
getArgument (Tcl_ThreadId threadId,
             Tcl_Interp *interp,
             va_list pArg,
             TclObject &tclObject,
             const Parameter &param)
{
    if (param.flags() & PARAMFLAG_FOUT) {
        // Get name of Tcl variable to hold out value.
        TclObject varName = getOutVariableName(param);

        if (param.flags() & PARAMFLAG_FIN) {
            // For in/out parameters, set the Tcl variable to the input value.
            TclObject value;
            pArg = convertNativeToTclObject(
                pArg, interp, value, param.type(), true);

            // Construct a Tcl command to set the variable.
            TclObject script(Tcl_NewListObj(0, 0));
            script.lappend(Tcl_NewStringObj("set", -1));
            script.lappend(varName);
            script.lappend(value);

            CommandProcessor::instance().eval(threadId, interp, script);
        } else {
            // Advance to next argument.
            va_arg(pArg, void *);
        }
        tclObject = varName;
        return pArg;

    } else {
        return convertNativeToTclObject(
            pArg, interp, tclObject, param.type());
    }
}

// Convert Tcl value to native value and store it at the address the va_list
// points to.
// Returns a va_list pointing to the next argument.

static va_list
putArgument (va_list pArg,
             Tcl_Interp *interp,
             TclObject tclObject,
             const Type &type)
{
    void *pDest = va_arg(pArg, void *);
    if (pDest == 0) {
        return pArg;
    }

    switch (type.vartype()) {
    case VT_BOOL:
        *reinterpret_cast<VARIANT_BOOL *>(pDest) =
            tclObject.getBool() ? VARIANT_TRUE : VARIANT_FALSE;
        break;

    case VT_R4:
        *reinterpret_cast<float *>(pDest) = tclObject.getDouble();
        break;

    case VT_R8:
        *reinterpret_cast<double *>(pDest) = tclObject.getDouble();
        break;

    case VT_DISPATCH:
    case VT_UNKNOWN:
    case VT_USERDEFINED:
        {
            IUnknown *pUnknown;

            Tcl_Obj *pObj = tclObject;
            if (pObj->typePtr == &unknownPointerType) {
                pUnknown = reinterpret_cast<IUnknown *>(
                    pObj->internalRep.otherValuePtr);
            } else {
                Reference *pRef = referenceHandleSupport.find(tclObject);
                pUnknown = (pRef == 0) ? 0 : pRef->unknown();
            }

            *reinterpret_cast<IUnknown **>(pDest) = pUnknown;

            // The COM rules say we must increment the reference count of
            // interface pointers returned from methods.
            if (pUnknown != 0) {
                pUnknown->AddRef();
            }
        }
        break;

    case VT_BSTR:
        {
            BSTR *pBstr = reinterpret_cast<BSTR *>(pDest);
#if TCL_MINOR_VERSION >= 2
            // Uses Unicode function introduced in Tcl 8.2.
            *pBstr = SysAllocString(tclObject.getUnicode());
#else
            *pBstr = SysAllocString(tclObject.c_str());
#endif
        }
        break;

    case VT_VARIANT:
        {
            // Must increment reference count of interface pointers returned
            // from methods.
            tclObject.toVariant(
                reinterpret_cast<VARIANT *>(pDest),
                Type::variant(),
                interp,
                true);
        }
        break;

    default:
        *reinterpret_cast<int *>(pDest) = tclObject.getLong();
    }

    return pArg;
}

// Advance the va_list to the next argument.
// Returns a va_list pointing to the next argument.

static va_list
nextArgument (va_list pArg, const Type &type)
{
    switch (type.vartype()) {
    case VT_R4:
    case VT_R8:
    case VT_DATE:
        va_arg(pArg, double);
        break;

    case VT_DISPATCH:
    case VT_UNKNOWN:
    case VT_USERDEFINED:
        va_arg(pArg, IUnknown *);
        break;

    case VT_BSTR:
        va_arg(pArg, BSTR);
        break;

    default:
        va_arg(pArg, int);
    }

    return pArg;
}

// Set error info.

static void
setErrorInfo (const char *source, const char *description)
{
    HRESULT hr;

    ICreateErrorInfoPtr pCreateErrorInfo;
    hr = CreateErrorInfo(&pCreateErrorInfo);
    if (FAILED(hr)) {
        return;
    }

    _bstr_t sourceBstr(source);
    pCreateErrorInfo->SetSource(sourceBstr);

    _bstr_t descriptionBstr(description);
    pCreateErrorInfo->SetDescription(descriptionBstr);

    IErrorInfoPtr pErrorInfo;
    hr = pCreateErrorInfo->QueryInterface(
        IID_IErrorInfo, reinterpret_cast<void **>(&pErrorInfo));
    if (SUCCEEDED(hr)) {
        SetErrorInfo(0, pErrorInfo);
    }
}

// Note that this function is called in an odd way to avoid copying the
// arguments onto the stack (for efficiency and simplicity in the calling
// code).  This is why the call is explicitly declared __cdecl.

void __cdecl
invokeComObjectFunction (volatile HRESULT hresult,
                         volatile DWORD pArgEnd,
	                 DWORD /*ebp*/,
	                 DWORD funcIndex,
	                 DWORD /*retAddr*/,
	                 InterfaceAdapter *pAdapter,
                         ...)
{
    // Get the method description for method being invoked.
    const Method *pMethod = pAdapter->findComMethod(funcIndex);
    if (pMethod == 0) {
        // If we don't have a method description, we don't know how many bytes
        // the arguments take on the stack.
        throw std::runtime_error("unknown virtual function index");
    }

    ComObject &object = pAdapter->object();

    // Construct Tcl command call.
    TclObject script(object.m_servant);

    std::string operation;
    switch (pMethod->invokeKind()) {
    case INVOKE_PROPERTYGET:
        operation = getPrefix + pMethod->name();
        break;
    case INVOKE_PROPERTYPUT:
    case INVOKE_PROPERTYPUTREF:
        operation = setPrefix + pMethod->name();
        break;
    default:
        operation = pMethod->name();
    }
    script.lappend(
        Tcl_NewStringObj(const_cast<char *>(operation.c_str()), -1));

    // Convert arguments to Tcl values.
    va_list pArg;
    va_start(pArg, pAdapter);
    const Method::Parameters &parameters = pMethod->parameters();
    Method::Parameters::const_iterator pParam = parameters.begin();
    for (; pParam != parameters.end(); ++pParam) {
        // Append argument value.
        TclObject argument;
        pArg = getArgument(
            object.m_threadId, object.m_interp, pArg, argument, *pParam);
        script.lappend(argument);
    }

    // Set end of arguments pointer.
    if (pMethod->type().vartype() != VT_VOID) {
        va_arg(pArg, void *);
    }
    pArgEnd = reinterpret_cast<DWORD>(pArg);
    va_end(pArg);

    // Call the Tcl command.
    TclObject result;
    int completionCode = object.eval(script, object.m_isSink ? 0 : &result);
    if (completionCode != TCL_OK) {
        setErrorInfo(object.m_servant.c_str(), result.c_str());
        hresult = E_UNEXPECTED;
        return;
    }

    // Get values of out parameters.
    va_start(pArg, pAdapter);
    pParam = parameters.begin();
    for (; pParam != parameters.end(); ++pParam) {
        if (pParam->flags() & PARAMFLAG_FOUT) {
            // Get name of Tcl variable that holds out value.
            TclObject varName = getOutVariableName(*pParam);

            // Construct a Tcl command to get the variable value.
            TclObject script(Tcl_NewListObj(0, 0));
            script.lappend(Tcl_NewStringObj("set", -1));
            script.lappend(varName);

            TclObject value;
            object.eval(script, &value);

            pArg = putArgument(pArg, object.m_interp, value, pParam->type());
        } else {
            pArg = nextArgument(pArg, pParam->type());
        }
    }

    // Convert return value.
    if (pMethod->type().vartype() != VT_VOID) {
        putArgument(pArg, object.m_interp, result, pMethod->type());
    }

    va_end(pArg);
    hresult = S_OK;
}
