// $Id: TclObject.cpp,v 1.22 2001/08/21 01:41:52 cthuang Exp $
#include "TclObject.h"
#ifdef WIN32
#include "refCmd.h"
#include "naCmd.h"
#include "nullCmd.h"
#include "objectCmd.h"
#endif

TclObject::TclObject ():
    m_pObj(Tcl_NewObj())
{ Tcl_IncrRefCount(m_pObj); }

TclObject::TclObject (const TclObject &rhs):
    m_pObj(rhs.m_pObj)
{ Tcl_IncrRefCount(m_pObj); }

TclObject::TclObject (Tcl_Obj *pObj):
    m_pObj(pObj)
{ Tcl_IncrRefCount(m_pObj); }

TclObject::TclObject (const char *src, int len):
    m_pObj(Tcl_NewStringObj(src, len))
{ Tcl_IncrRefCount(m_pObj); }

#if TCL_MINOR_VERSION >= 2
TclObject::TclObject (const wchar_t *src, int len):
    m_pObj(Tcl_NewUnicodeObj(
	const_cast<Tcl_UniChar *>(reinterpret_cast<const Tcl_UniChar *>(src)),
	len))
{ Tcl_IncrRefCount(m_pObj); }
#endif

TclObject::TclObject (const std::string &s):
    m_pObj(Tcl_NewStringObj(s.data(), s.size()))
{ Tcl_IncrRefCount(m_pObj); }

TclObject::TclObject (bool value):
    m_pObj(Tcl_NewBooleanObj(static_cast<int>(value)))
{ Tcl_IncrRefCount(m_pObj); }

TclObject::TclObject (int value):
    m_pObj(Tcl_NewIntObj(value))
{ Tcl_IncrRefCount(m_pObj); }

TclObject::TclObject (long value):
    m_pObj(Tcl_NewLongObj(value))
{ Tcl_IncrRefCount(m_pObj); }

TclObject::TclObject (double value):
    m_pObj(Tcl_NewDoubleObj(value))
{ Tcl_IncrRefCount(m_pObj); }

TclObject::TclObject (int objc, Tcl_Obj *CONST objv[]):
    m_pObj(Tcl_NewListObj(objc, objv))
{ Tcl_IncrRefCount(m_pObj); }

TclObject::~TclObject ()
{ Tcl_DecrRefCount(m_pObj); }

TclObject &
TclObject::operator= (const TclObject &rhs)
{
    Tcl_IncrRefCount(rhs.m_pObj);
    Tcl_DecrRefCount(m_pObj);
    m_pObj = rhs.m_pObj;
    return *this;
}

TclObject &
TclObject::operator= (Tcl_Obj *pObj)
{
    Tcl_IncrRefCount(pObj);
    Tcl_DecrRefCount(m_pObj);
    m_pObj = pObj;
    return *this;
}

bool
TclObject::getBool () const
{
    int value;
    Tcl_GetBooleanFromObj(0, m_pObj, &value);
    return value != 0;
}

int
TclObject::getInt () const
{
    int value;
    Tcl_GetIntFromObj(0, m_pObj, &value);
    return value;
}

long
TclObject::getLong () const
{
    long value;
    Tcl_GetLongFromObj(0, m_pObj, &value);
    return value;
}

double
TclObject::getDouble () const
{
    double value;
    Tcl_GetDoubleFromObj(0, m_pObj, &value);
    return value;
}

TclObject &
TclObject::lappend (Tcl_Obj *pElement)
{
    if (Tcl_IsShared(m_pObj)) {
        Tcl_DecrRefCount(m_pObj);
        m_pObj = Tcl_DuplicateObj(m_pObj);
        Tcl_IncrRefCount(m_pObj);
    }
    Tcl_ListObjAppendElement(NULL, m_pObj, pElement);
    // TODO: Should check for error result if conversion to list failed.
    return *this;
}

#ifdef WIN32

static Tcl_Obj *
newUnicodeObj (const Tcl_UniChar *pWide, int length)
{
    if (pWide == 0) {
        return Tcl_NewObj();
    }
    return Tcl_NewUnicodeObj(const_cast<Tcl_UniChar *>(pWide), length);
}

TclObject::TclObject (VARIANT *pSrc, const Type &type, Tcl_Interp *interp)
{
    HRESULT hr;

    if (V_VT(pSrc) & VT_ARRAY) {
        // We can handle only one-dimensional arrays.
        SAFEARRAY *psa = V_ARRAY(pSrc);
        if (SafeArrayGetDim(psa) != 1) {
            _com_issue_error(E_INVALIDARG);
        }

        // Get index range.
        long lowerBound;
        hr = SafeArrayGetLBound(psa, 1, &lowerBound);
        if (FAILED(hr)) {
            _com_issue_error(hr);
        }

        long upperBound;
        hr = SafeArrayGetUBound(psa, 1, &upperBound);
        if (FAILED(hr)) {
            _com_issue_error(hr);
        }

        // Get element type.
        VARTYPE vt = V_VT(pSrc) & VT_TYPEMASK;

        switch (vt) {
        case VT_UNKNOWN:
            // Convert array of IUnknown to Tcl list of interface pointer
            // handles.
            {
                IUnknown **pData;
                hr = SafeArrayAccessData(
                    psa, reinterpret_cast<void **>(&pData));
                if (FAILED(hr)) {
                    _com_issue_error(hr);
                }

                m_pObj = Tcl_NewListObj(0, 0);
                for (long i = lowerBound; i <= upperBound; ++i) {
                    Tcl_Obj *pElement =
                        referenceHandleSupport.newObj(
                            interp, Reference::newReference(pData[i]));
                    Tcl_ListObjAppendElement(interp, m_pObj, pElement);
                }

                hr = SafeArrayUnaccessData(psa);
                if (FAILED(hr)) {
                    _com_issue_error(hr);
                }
            }
            break;

        case VT_UI1:
            {
                unsigned char *pData;
                hr = SafeArrayAccessData(
                    psa, reinterpret_cast<void **>(&pData));
                if (FAILED(hr)) {
                    _com_issue_error(hr);
                }

                long length = upperBound - lowerBound + 1;
                m_pObj =
#if TCL_MINOR_VERSION >= 1
                    // Convert array of bytes to Tcl byte array.
                    Tcl_NewByteArrayObj(pData, length);
#else
                    // Convert array of bytes to Tcl string.
                    Tcl_NewStringObj(reinterpret_cast<char *>(pData), length);
#endif

                hr = SafeArrayUnaccessData(psa);
                if (FAILED(hr)) {
                    _com_issue_error(hr);
                }
            }
            break;

        default:
            // Convert array of other types to Tcl list of objects.
            {
                m_pObj = Tcl_NewListObj(0, 0);
                for (long i = lowerBound; i <= upperBound; ++i) {
                    _variant_t elementVar;

                    if (vt == VT_VARIANT) {
                        hr = SafeArrayGetElement(psa, &i, &elementVar);
                    } else {
                        // I hope the element can be contained in a VARIANT.
                        elementVar.ChangeType(vt);
                        hr = SafeArrayGetElement(
                            psa,
                            &i,
                            &elementVar.punkVal);
                    }
                    if (FAILED(hr)) {
                        _com_issue_error(hr);
                    }

                    TclObject element(&elementVar, type, interp);
                    Tcl_ListObjAppendElement(interp, m_pObj, element);
                }
            }
        }

    } else {
        switch (V_VT(pSrc)) {
        case VT_BOOL:
            m_pObj = Tcl_NewBooleanObj(V_BOOL(pSrc));
            break;

        case VT_I1:
        case VT_UI1:
            m_pObj = Tcl_NewLongObj(V_I1(pSrc));
            break;

        case VT_I2:
        case VT_UI2:
            m_pObj = Tcl_NewLongObj(V_I2(pSrc));
            break;

        case VT_I4:
        case VT_UI4:
        case VT_INT:
        case VT_UINT:
            m_pObj = Tcl_NewLongObj(V_I4(pSrc));
            break;

        case VT_R4:
            m_pObj = Tcl_NewDoubleObj(V_R4(pSrc));
            break;

        case VT_DATE:
        case VT_R8:
            m_pObj = Tcl_NewDoubleObj(V_R8(pSrc));
            break;

        case VT_DISPATCH:
            {
                const Interface *pInterface =
                    InterfaceManager::instance().find(type.iid());
                m_pObj = referenceHandleSupport.newObj(
                    interp,
                    Reference::newReference(V_DISPATCH(pSrc), pInterface));
            }
            break;

        case VT_UNKNOWN:
            {
                const Interface *pInterface =
                    InterfaceManager::instance().find(type.iid());
                m_pObj = referenceHandleSupport.newObj(
                    interp,
                    Reference::newReference(V_UNKNOWN(pSrc), pInterface));
            }
            break;

        case VT_NULL:
            m_pObj = Tcl_NewObj();
            break;

        case VT_LPSTR:
            m_pObj = Tcl_NewStringObj(V_I1REF(pSrc), -1);
            break;

        case VT_LPWSTR:
            {
#if TCL_MINOR_VERSION >= 2
                // Uses Unicode function introduced in Tcl 8.2.
                m_pObj = newUnicodeObj(V_UI2REF(pSrc), -1);
#else
		const wchar_t *pWide = V_UI2REF(pSrc);
                _bstr str(pWide);
                m_pObj = Tcl_NewStringObj(str, -1);
#endif
            }
            break;

        default:
            if (V_VT(pSrc) == VT_USERDEFINED && type.name() == "GUID") {
                Uuid uuid(*static_cast<UUID *>(V_BYREF(pSrc)));
                m_pObj = Tcl_NewStringObj(uuid.toString().c_str(), -1);
            } else {
                _bstr_t str(pSrc);
#if TCL_MINOR_VERSION >= 2
                // Uses Unicode function introduced in Tcl 8.2.
		wchar_t *pWide = str;
                m_pObj = newUnicodeObj(
                    reinterpret_cast<Tcl_UniChar *>(pWide), str.length());
#else
                m_pObj = Tcl_NewStringObj(str, -1);
#endif
            }
        }
    }

    Tcl_IncrRefCount(m_pObj);
}

BSTR
TclObject::getBSTR () const
{
#if TCL_MINOR_VERSION >= 2
    // Uses Unicode function introduced in Tcl 8.2.
    return SysAllocString(getUnicode());
#else
    _bstr_t str(c_str());
    return SysAllocString(str);
#endif
}

void
TclObject::toVariant (VARIANT *pDest,
                      const Type &type,
                      Tcl_Interp *interp,
                      bool addRef)
{
    VariantClear(pDest);
    VARTYPE vt = type.vartype();

    if (referenceHandleSupport.count(m_pObj) > 0) {
        // Convert interface pointer handle to interface pointer.
        Reference *pReference = referenceHandleSupport.find(m_pObj);
        if (addRef) {
            // Must increment reference count of interface pointers returned
            // from methods.
            pReference->unknown()->AddRef();
        }

        IDispatch *pDispatch = pReference->dispatch();
        if (pDispatch != 0) {
            V_VT(pDest) = VT_DISPATCH;
            V_DISPATCH(pDest) = pDispatch;
        } else {
            V_VT(pDest) = VT_UNKNOWN;
            V_UNKNOWN(pDest) = pReference->unknown();
        }

    } else if (m_pObj->typePtr == &unknownPointerType) {
        // Convert to interface pointer.
        IUnknown *pUnknown = static_cast<IUnknown *>(
            m_pObj->internalRep.otherValuePtr);
        if (addRef) {
            // Must increment reference count of interface pointers returned
            // from methods.
            pUnknown->AddRef();
        }
        V_VT(pDest) = VT_UNKNOWN;
        V_UNKNOWN(pDest) = pUnknown;

    } else if (vt == VT_SAFEARRAY) {
        // Convert Tcl list to SAFEARRAY.
        int numElements;
        Tcl_Obj **pElements;
        if (Tcl_ListObjGetElements(interp, m_pObj, &numElements, &pElements)
          != TCL_OK) {
            _com_issue_error(E_INVALIDARG);
        }

        const Type &elementType = type.elementType();
        SAFEARRAY *psa =
            SafeArrayCreateVector(elementType.vartype(), 0, numElements);
        if (psa == 0) {
            _com_issue_error(E_OUTOFMEMORY);
        }

        void *pData;
        HRESULT hr;
        hr = SafeArrayAccessData(psa, &pData);
        if (FAILED(hr)) {
            _com_issue_error(hr);
        }

        for (int i = 0; i < numElements; ++i) {
            TclObject value(pElements[i]);

            switch (elementType.vartype()) {
            case VT_BOOL:
                static_cast<VARIANT_BOOL *>(pData)[i] =
                    value.getBool() ? VARIANT_TRUE : VARIANT_FALSE;
                break;

            case VT_R4:
                static_cast<float *>(pData)[i] = value.getDouble();
                break;

            case VT_R8:
                static_cast<double *>(pData)[i] = value.getDouble();
                break;

            case VT_BSTR:
                static_cast<BSTR *>(pData)[i] = value.getBSTR();
                break;

            case VT_VARIANT:
                {
                    VARIANT *pDest = static_cast<VARIANT *>(pData) + i;
                    VariantInit(pDest);
                    value.toVariant(pDest, elementType, interp);
                }
                break;

            default:
                static_cast<int *>(pData)[i] = value.getLong();
            }
        }

        hr = SafeArrayUnaccessData(psa);
        if (FAILED(hr)) {
            _com_issue_error(hr);
        }

        V_VT(pDest) = VT_ARRAY | elementType.vartype();
        V_ARRAY(pDest) = psa;

    } else if (m_pObj->typePtr == Tcl_GetObjType("list")) {
        // Convert Tcl list to array of VARIANT.
        int numElements;
        Tcl_Obj **pElements;
        if (Tcl_ListObjGetElements(interp, m_pObj, &numElements, &pElements)
          != TCL_OK) {
            _com_issue_error(E_INVALIDARG);
        }

        SAFEARRAY *psa = SafeArrayCreateVector(VT_VARIANT, 0, numElements);
        if (psa == 0) {
            _com_issue_error(E_OUTOFMEMORY);
        }

        VARIANT *pData;
        HRESULT hr;
        hr = SafeArrayAccessData(psa, reinterpret_cast<void **>(&pData));
        if (FAILED(hr)) {
            _com_issue_error(hr);
        }

        for (int i = 0; i < numElements; ++i) {
            TclObject value(pElements[i]);
            VariantInit(&pData[i]);
            value.toVariant(&pData[i], Type::variant(), interp, addRef);
        }

        hr = SafeArrayUnaccessData(psa);
        if (FAILED(hr)) {
            _com_issue_error(hr);
        }

        V_VT(pDest) = VT_ARRAY | VT_VARIANT;
        V_ARRAY(pDest) = psa;

#if TCL_MINOR_VERSION >= 1
    } else if (m_pObj->typePtr == Tcl_GetObjType("bytearray")) {
        // Convert Tcl byte array to array of bytes.
        int length;
        unsigned char *pBytes = Tcl_GetByteArrayFromObj(m_pObj, &length);

        SAFEARRAY *psa = SafeArrayCreateVector(VT_UI1, 0, length);
        if (psa == 0) {
            _com_issue_error(E_OUTOFMEMORY);
        }

        unsigned char *pDestData;
        HRESULT hr;
        hr = SafeArrayAccessData(
            psa, reinterpret_cast<void **>(&pDestData));
        if (FAILED(hr)) {
            _com_issue_error(hr);
        }

        memcpy(pDestData, pBytes, length);

        hr = SafeArrayUnaccessData(psa);
        if (FAILED(hr)) {
            _com_issue_error(hr);
        }

        V_VT(pDest) = VT_ARRAY | VT_UI1;
        V_ARRAY(pDest) = psa;
#endif

    } else if (m_pObj->typePtr == &naType) {
        // This variant indicates a missing optional argument.
        VariantCopy(pDest, &vtMissing);

    } else if (m_pObj->typePtr == &nullType) {
        V_VT(pDest) = VT_NULL;

    } else if (m_pObj->typePtr == Tcl_GetObjType("int")) {
        long value;
        if (Tcl_GetLongFromObj(interp, m_pObj, &value) != TCL_OK) {
            value = 0;
        }
        V_VT(pDest) = VT_I4;
        V_I4(pDest) = value;

        if (vt != VT_VARIANT && vt != VT_USERDEFINED) {
            VariantChangeType(pDest, pDest, 0, vt);
        }

    } else if (m_pObj->typePtr == Tcl_GetObjType("double")) {
        double value;
        if (Tcl_GetDoubleFromObj(interp, m_pObj, &value) != TCL_OK) {
            value = 0.0;
        }
        V_VT(pDest) = VT_R8;
        V_R8(pDest) = value;

        if (vt != VT_VARIANT && vt != VT_USERDEFINED) {
            VariantChangeType(pDest, pDest, 0, vt);
        }

    } else if (m_pObj->typePtr == Tcl_GetObjType("boolean")) {
        int value;
        if (Tcl_GetBooleanFromObj(interp, m_pObj, &value) != TCL_OK) {
            value = 0;
        }
        V_VT(pDest) = VT_BOOL;
        V_BOOL(pDest) = (value != 0) ? VARIANT_TRUE : VARIANT_FALSE;

        if (vt != VT_VARIANT && vt != VT_USERDEFINED) {
            VariantChangeType(pDest, pDest, 0, vt);
        }

    } else {
#if TCL_MINOR_VERSION >= 2
        // Uses Unicode function introduced in Tcl 8.2.
        const wchar_t *pStringRep =
	    reinterpret_cast<const wchar_t *>(Tcl_GetUnicode(m_pObj));
#else
        const char *pStringRep = Tcl_GetStringFromObj(m_pObj, 0);
#endif
        _variant_t var(pStringRep);

        // If trying to convert from a string to a date,
        // we need to convert to a double (VT_R8) first.
        if (vt == VT_DATE) {
            var.ChangeType(VT_R8);
        }

        // Try to convert from a string representation.
        if (vt != VT_VARIANT && vt != VT_USERDEFINED && vt != VT_LPWSTR) {
            var.ChangeType(vt);
        }
        VariantCopy(pDest, &var);
    }
}

#endif
